@@ -4,16 +4,16 @@
 import collections
 from functools import partial
 from itertools import product
-from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Tuple, Union
+from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union
 
 import pywt
 import torch
 
-
 from ._util import Wavelet, _as_wavelet
-from .conv_transform import wavedec, wavedec2
+from .conv_transform import wavedec
+from .conv_transform_2 import wavedec2
 from .matmul_transform import MatrixWavedec
-from .matmul_transform_2d import MatrixWavedec2d
+from .matmul_transform_2 import MatrixWavedec2
 
 if TYPE_CHECKING:
     BaseDict = collections.UserDict[str, torch.Tensor]
@@ -30,6 +30,7 @@ class WaveletPacket(BaseDict):
         wavelet: Union[Wavelet, str],
         mode: str = "reflect",
         boundary_orthogonalization: str = "qr",
+        max_level: Optional[int] = None,
     ) -> None:
         """Create a wavelet packet decomposition object.
 
@@ -46,17 +47,20 @@ class WaveletPacket(BaseDict):
             boundary_orthogonalization (str): The orthogonalization method
                 to use. Only used if `mode` equals 'boundary'. Choose from
                 'qr' or 'gramschmidt'. Defaults to 'qr'.
+            max_level (int, optional): Value is passed on to `transform`.
+                The highest decomposition level to compute. If None, the maximum level
+                is determined from the input data shape. Defaults to None.
         """
         self.wavelet = _as_wavelet(wavelet)
-        if mode == "zero":
-            # translate pywt to PyTorch
-            self.mode = "constant"
-        else:
-            self.mode = mode
+        self.mode = mode
         self.boundary = boundary_orthogonalization
         self._matrix_wavedec_dict: Dict[int, MatrixWavedec] = {}
+        self.max_level: Optional[int] = None
         if data is not None:
-            self.transform(data)
+            if len(data.shape) == 1:
+                # add a batch dimension.
+                data = data.unsqueeze(0)
+            self.transform(data, max_level)
         else:
             self.data = {}
 
@@ -111,22 +115,47 @@ class WaveletPacket(BaseDict):
             ]
         return graycode_order
 
-    # ignoring missing return type, as recursive nesting is currently not supported
-    # see https://github.com/python/mypy/issues/731
-    def _recursive_dwt(  # type: ignore[no-untyped-def]
-        self, data: torch.Tensor, level: int, path: str
-    ):
+    def _recursive_dwt(self, data: torch.Tensor, level: int, path: str) -> None:
         if not self.max_level:
             raise AssertionError
-        self.data[path] = torch.squeeze(data)
+
+        # TODO: This is a workaround since the convolutional transforms insert a
+        #       squeezable dimension. We should adapt the wavedec code instead.
+        if data.dim() == 3:
+            data = data.squeeze(1)
+
+        self.data[path] = data
         if level < self.max_level:
             res_lo, res_hi = self._get_wavedec(data.shape[-1])(data)
-            return (
-                self._recursive_dwt(res_lo, level + 1, path + "a"),
-                self._recursive_dwt(res_hi, level + 1, path + "d"),
+            self._recursive_dwt(res_lo, level + 1, path + "a")
+            self._recursive_dwt(res_hi, level + 1, path + "d")
+
+    def __getitem__(self, key: str) -> torch.Tensor:
+        """Access the coefficients in the wavelet packets tree.
+
+        Args:
+            key (str): The key of the accessed coefficents. The string may only consist
+                of the following chars: 'a', 'd'.
+
+        Returns:
+            torch.Tensor: The accessed wavelet packet coefficients.
+
+        Raises:
+            ValueError: If the wavelet packet tree is not initialized.
+            KeyError: If no wavelet coefficients are indexed by the specified key.
+        """
+        if self.max_level is None:
+            raise ValueError(
+                "The wavelet packet tree must be initialized via 'transform' before "
+                "its values can be accessed!"
             )
-        else:
-            self.data[path] = torch.squeeze(data)
+        if key not in self and len(key) > self.max_level:
+            raise KeyError(
+                f"The requested level {len(key)} with key '{key}' is too large and "
+                "cannot be accessed! This wavelet packet tree is initialized with "
+                f"maximum level {self.max_level}."
+            )
+        return super().__getitem__(key)
 
 
 class WaveletPacket2D(BaseDict):
@@ -139,6 +168,7 @@ class WaveletPacket2D(BaseDict):
         mode: str = "reflect",
         boundary_orthogonalization: str = "qr",
         separable: bool = False,
+        max_level: Optional[int] = None,
     ) -> None:
         """Create a 2D-Wavelet packet tree.
 
@@ -158,19 +188,19 @@ class WaveletPacket2D(BaseDict):
             separable (bool): If true and the sparse matrix backend is selected,
                 a separable transform is performed, i.e. each image axis is
                 transformed separately. Defaults to False.
+            max_level (int, optional): Value is passed on to `transform`.
+                The highest decomposition level to compute. If None, the maximum level
+                is determined from the input data shape. Defaults to None.
         """
         self.wavelet = _as_wavelet(wavelet)
-
-        # translate pywt to PyTorch
-        self.mode = "constant" if mode == "zero" else mode
-
+        self.mode = mode
         self.boundary = boundary_orthogonalization
         self.separable = separable
-        self.matrix_wavedec2_dict: Dict[Tuple[int, ...], MatrixWavedec2d] = {}
+        self.matrix_wavedec2_dict: Dict[Tuple[int, ...], MatrixWavedec2] = {}
 
         self.max_level: Optional[int] = None
         if data is not None:
-            self.transform(data)
+            self.transform(data, max_level)
         else:
             self.data = {}
 
@@ -193,9 +223,6 @@ class WaveletPacket2D(BaseDict):
             max_level = pywt.dwt_max_level(min(data.shape[-2:]), self.wavelet.dec_len)
         self.max_level = max_level
 
-        if data.dim() == 3:
-            data = torch.unsqueeze(data, 1)
-
         self._recursive_dwt2d(data, level=0, path="")
         return self
 
@@ -208,7 +235,7 @@ class WaveletPacket2D(BaseDict):
         if self.mode == "boundary":
             shape = tuple(shape)
             if shape not in self.matrix_wavedec2_dict.keys():
-                self.matrix_wavedec2_dict[shape] = MatrixWavedec2d(
+                self.matrix_wavedec2_dict[shape] = MatrixWavedec2(
                     self.wavelet,
                     level=1,
                     boundary=self.boundary,
@@ -219,36 +246,60 @@ class WaveletPacket2D(BaseDict):
         else:
             return partial(wavedec2, wavelet=self.wavelet, level=1, mode=self.mode)
 
-    # ignoring missing return type, as recursive nesting is currently not supported
-    # see https://github.com/python/mypy/issues/731
-    def _recursive_dwt2d(  # type: ignore[no-untyped-def]
-        self, data: torch.Tensor, level: int, path: str
-    ):
+    def _recursive_dwt2d(self, data: torch.Tensor, level: int, path: str) -> None:
         if not self.max_level:
             raise AssertionError
+
+        # TODO: This is a workaround since the convolutional transforms insert a
+        #       squeezable dimension. We should adapt the wavedec2 code instead.
+        if data.dim() == 4:
+            data = data.squeeze(1)
+
         self.data[path] = data
         if level < self.max_level:
-            # resa, (resh, resv, resd) = self.wavedec2(
-            #    data, self.wavelet, level=1, mode=self.mode
-            # )
             result_a, (result_h, result_v, result_d) = self._get_wavedec(
                 data.shape[-2:]
             )(data)
             # assert for type checking
             assert not isinstance(result_a, tuple)
-            return (
-                self._recursive_dwt2d(result_a, level + 1, path + "a"),
-                self._recursive_dwt2d(result_h, level + 1, path + "h"),
-                self._recursive_dwt2d(result_v, level + 1, path + "v"),
-                self._recursive_dwt2d(result_d, level + 1, path + "d"),
+            self._recursive_dwt2d(result_a, level + 1, path + "a")
+            self._recursive_dwt2d(result_h, level + 1, path + "h")
+            self._recursive_dwt2d(result_v, level + 1, path + "v")
+            self._recursive_dwt2d(result_d, level + 1, path + "d")
+
+    def __getitem__(self, key: str) -> torch.Tensor:
+        """Access the coefficients in the wavelet packets tree.
+
+        Args:
+            key (str): The key of the accessed coefficents. The string may only consist
+                of the following chars: 'a', 'h', 'v', 'd'.
+
+        Returns:
+            torch.Tensor: The accessed wavelet packet coefficients.
+
+        Raises:
+            ValueError: If the wavelet packet tree is not initialized.
+            KeyError: If no wavelet coefficients are indexed by the specified key.
+        """
+        if self.max_level is None:
+            raise ValueError(
+                "The wavelet packet tree must be initialized via 'transform' before "
+                "its values can be accessed!"
             )
-        else:
-            self.data[path] = torch.squeeze(data)
+        if key not in self and len(key) > self.max_level:
+            raise KeyError(
+                f"The requested level {len(key)} with key '{key}' is too large and "
+                "cannot be accessed! This wavelet packet tree is initialized with "
+                f"maximum level {self.max_level}."
+            )
+        return super().__getitem__(key)
 
 
 def get_freq_order(level: int) -> List[List[Tuple[str, ...]]]:
     """Get the frequency order for a given packet decomposition level.
 
+    Use this code to create two-dimensional frequency orderings.
+
     Args:
         level (int): The number of decomposition scales.
 
@@ -276,7 +327,7 @@ def get_freq_order(level: int) -> List[List[Tuple[str, ...]]]:
             ]
         return graycode_order
 
-    def expand_2d_path(path: Tuple[str, ...]) -> Tuple[str, str]:
+    def _expand_2d_path(path: Tuple[str, ...]) -> Tuple[str, str]:
         expanded_paths = {"d": "hh", "h": "hl", "v": "lh", "a": "ll"}
         return (
             "".join([expanded_paths[p][0] for p in path]),
@@ -285,7 +336,7 @@ def get_freq_order(level: int) -> List[List[Tuple[str, ...]]]:
 
     nodes_dict: Dict[str, Dict[str, Tuple[str, ...]]] = {}
     for (row_path, col_path), node in [
-        (expand_2d_path(node), node) for node in wp_natural_path
+        (_expand_2d_path(node), node) for node in wp_natural_path
     ]:
         nodes_dict.setdefault(row_path, {})[col_path] = node
     graycode_order = _get_graycode_order(level, x="l", y="h")
@@ -294,93 +345,3 @@ def get_freq_order(level: int) -> List[List[Tuple[str, ...]]]:
     for row in nodes:
         result.append([row[path] for path in graycode_order if path in row])
     return result
-
-
-if __name__ == "__main__":
-    import numpy as np
-
-    # import matplotlib.pyplot as plt
-    # import scipy.signal as signal
-    # from itertools import product
-
-    from scipy import misc
-
-    face = misc.face()[:128, :128]
-    wavelet = pywt.Wavelet("haar")
-    wp_tree = pywt.WaveletPacket2D(
-        data=np.mean(face, axis=-1).astype(np.float32), wavelet=wavelet, mode="zero"
-    )
-
-    # Get the full decomposition
-    max_lev = 5
-    wp_keys = list(product(["a", "d", "h", "v"], repeat=max_lev))
-    count = 0
-    img_rows = None
-    img = []
-    for node in wp_keys:
-        packet = np.squeeze(wp_tree["".join(node)].data)
-        if img_rows is not None:
-            img_rows = np.concatenate([img_rows, packet], axis=1)
-        else:
-            img_rows = packet
-        count += 1
-        if count > 31:
-            count = 0
-            img.append(img_rows)
-            img_rows = None
-
-    img_pywt = np.concatenate(img, axis=0)
-
-    pt_data = torch.unsqueeze(
-        torch.from_numpy(np.mean(face, axis=-1).astype(np.float32)), 0
-    )
-    pt_data = torch.cat([pt_data, pt_data], 0)
-    ptwt_wp_tree = WaveletPacket2D(data=pt_data, wavelet=wavelet, mode="boundary")
-
-    # get the PyTorch decomposition
-    count = 0
-    img_pt = []
-    img_rows_pt = None
-    for node in wp_keys:
-        packet = torch.squeeze(ptwt_wp_tree["".join(node)][0])
-        if img_rows_pt is not None:
-            img_rows_pt = torch.cat([img_rows_pt, packet], axis=1)
-        else:
-            img_rows_pt = packet
-        count += 1
-        if count > 31:
-            count = 0
-            img_pt.append(img_rows_pt)
-            img_rows_pt = None
-
-    img_pt = torch.cat(img_pt, dim=0).numpy()
-    abs = np.abs(img_pt - img_pywt)
-
-    err = np.mean(abs)
-    print("total error", err, ["ok" if err < 1e-4 else "failed!"])
-    assert err < 1e-4
-
-    print(
-        "a",
-        np.mean(
-            np.abs(wp_tree["a"].data - torch.squeeze(ptwt_wp_tree["a"][0]).numpy())
-        ),
-    )
-    print(
-        "h",
-        np.mean(
-            np.abs(wp_tree["h"].data - torch.squeeze(ptwt_wp_tree["h"][0]).numpy())
-        ),
-    )
-    print(
-        "v",
-        np.mean(
-            np.abs(wp_tree["v"].data - torch.squeeze(ptwt_wp_tree["v"][0]).numpy())
-        ),
-    )
-    print(
-        "d",
-        np.mean(
-            np.abs(wp_tree["d"].data - torch.squeeze(ptwt_wp_tree["d"][0]).numpy())
-        ),
-    )
